/**
* This is the browser-side javascript class for the DocSet viewer.
*
* It's implemented as a monolothic class to minimize any name-space pollution.
* 
* All the entry points are static methods toward the end. The actual stuff is done
* on a node server process, using sqlite3.
*
*/
class DClient {
	static client = new DClient({
		url: 'ws://localhost:8080/ws'
	});
	constructor(opt){
		this.prefix = opt.prefix || "";
		this.url = opt.url || 'ws://localhost:8080' + this.prefix + '/ws';
		this.is_closed = false;
		this.setup_websocket();
	}
	setup_websocket(){
		let me = this;
		let sock = new WebSocket(this.url);
		sock.addEventListener('open', (event) => { 
			me.is_closed = false;
		});
		sock.addEventListener("error", err => {
			console.log("OOPS: ",err);
			sock.close();
			me.is_closed = true;
			me.clear_options();
		});
		sock.addEventListener("close", ev => {
			me.is_closed = true;
			me.clear_options();
		});
		sock.addEventListener('message', (event)  => { 
			let rec = JSON.parse(event.data);
			switch(rec.data_type){
				case('hit'):
					this.rcv_hit(rec);
					break;
				case('id.type'):
					this.rcv_type(rec);
					break;
				case('id.collection'):
					this.rcv_collection(rec);
					break;
				case('hit-max'):
					alert(rec.title);
					break;
				default:
					console.log(`unknown ${rec.data_type}`);
			}
		});
		this.sock = sock;
	}

	// make the outer box for a group of hits. Example: "Vim" from vim docset.
	// the id of the UL will be "hit.list.Vim"
	create_hit_collection(hit){
		let list_id =  "hit.list." + hit.collection;
		let collection_id = "hit.group." + hit.collection;
		// where we're putting this collection
		let disp = document.getElementById("search.hits"); 

		// the outline of the collection.
		let tmpl = document.getElementById("hit.collection");
		let clone = tmpl.content.cloneNode(true);

		// label it.
		let a = clone.querySelector("a");
		a.innerHTML = "";
		a.appendChild(document.createTextNode(hit.collection));

		let q = clone.querySelector(".hit-collection-group");
		q.id = collection_id;

		let qr = clone.querySelector("ul.hide");
		qr.id = list_id;

		disp.appendChild(clone);
		return(qr);
	}
	rcv_type(hit){
		let tmpl = document.getElementById("input.type");
		let clone = tmpl.content.cloneNode(true);

		let input = clone.querySelector("input");
		input.value = hit.type;
		let label = clone.querySelector("label");
		label.appendChild(document.createTextNode(hit.type));

		let disp = document.getElementById("search.params.box.type.list");
		disp.append(clone);

	}
	clear_options(){
		let el = document.getElementById("search.params.box.collection.list");
		el.innerHTML = "";
		el = document.getElementById("search.params.box.type.list");
		el.innerHTML = "";
	}

	rcv_collection(row){
		let tmpl = document.getElementById("input.collection");
		let disp = document.getElementById("search.params.box.collection.list");
		row.collection.forEach( cname => {	
			let clone = tmpl.content.cloneNode(true);
			let input = clone.querySelector("input");
			input.value = cname;
			let label = clone.querySelector("label");
			label.appendChild(document.createTextNode(cname));

			disp.append(clone);
		});
	}
	/*
	* Resolve all the ../../ stuff and THEN make sure it is under 
	* the specified path.
	* 
	* Example:
	* "C","../../foo/bar/../car" -> "C/foo/car"
	*/
	grounded_path(gnd, p){
		// in case it's a <dash_embed junk>/path/to/file
		let ix = p.lastIndexOf(">");
		if(ix > -1){
			++ix;
			p = p.substr(ix);
		}
		let frag = null;
		ix = p.indexOf('#');
		if(ix > -1){
			frag = p.substr(ix);
			p = p.substr(0,ix);
		}
		let ar = p.split("/");
		let op = [];
		ar.forEach( el => {
			switch(el){
				case '..':
					op.shift();
					break;
				case '':
				case '.':
					break;
				default:
					op.push(el);
			}
		});
		if(frag){
			p = gnd + "/" + op.join("/") + frag;
		}else{
			p = gnd + "/" + op.join("/");
		}
		return(p);
	}
	/**
	* Called when a hit is received for the search results.
	*/
	rcv_hit(hit){
		let collection_id = "hit.list." + hit.collection;
		let disp = document.getElementById(collection_id);
		if(! disp){
			disp = this.create_hit_collection(hit);
		}
		let tmpl = document.getElementById("hit.entry");
		let clone = tmpl.content.cloneNode(true);
		let anchor = clone.querySelector("a.hit-entry");
		anchor.innerHTML = "";
		anchor.href="#" + this.grounded_path(hit.collection, hit.path);
		anchor.title = hit.collection + " :: " + hit.name + " (" + hit.type + ")";
		anchor.appendChild(document.createTextNode(hit.name));
		disp.append(clone);
	}

	do_toggle_collection_result(al){
		let id = "hit.list." + al.innerText;

		// Hide them all.
		let cel = document.getElementById("search.hits");
		let cl = cel.querySelectorAll("ul");
		cl.forEach( ul => {
			if(ul.id != id){
				ul.className = "hide";
			}
		});
		// then open the one we wanted.
		let el = document.getElementById(id);
		if(el){
			if(el.className == "hide"){
				el.className = "hit-item";
			}else{
				el.className = "hide";
			}
		}
	}

	do_fetch_doc(el){
		let ix = el.href.indexOf("#");
		if(ix < 0){
			return; // shouldn't happen.
		}
		let path = null;
		path = this.prefix + "/page/" + el.href.substr(ix + 1);
		let frame = document.getElementById("doc.box.frame");
		frame.src = path;
		let sel = document.getElementById("doc.box.select");

		let found = false;
		for( let ix = 0; ix < sel.length; ++ix){
			let op = sel.options[ix];
			if(op.value == path){
				sel.selectedIndex = ix;
				found = true;
			}
		}
		if(! found){
			let ix = sel.length;
			let op = document.createElement("option");
			op.text = el.title;
			op.value = path;
			sel.add(op);
			sel.selectedIndex = ix;
		}
		this.do_toggle_params("hide");
	}
	do_select_tab(sel){
		if(sel.options.length == 0){
			this.do_toggle_params(true);
			return;
		}
		let ix = sel.selectedIndex;
		let op = sel.options[ix];
		let frame = document.getElementById("doc.box.frame");
		frame.src = op.value;
		this.do_toggle_params("hide");
	}
	do_close_tab(el){
		let sel = document.getElementById("doc.box.select");
		let sel_ix = sel.selectedIndex;
		sel.remove(sel_ix);
		this.do_select_tab(sel);
	}
	do_toggle_params(state){
		let el = document.getElementById("search.params.box");
		let mb = document.getElementById("doc.box.main");
		if(state == "hide"){
			el.className = "hide";
			mb.className = "app-doc";
			return;
		}
		if(el.className == "hide"){
			// el.className = "search-params";
			el.className = "app-doc";
			mb.className = "hide";
		}else{
			el.className = "hide";
			mb.className = "app-doc";
		}
	}
	gather_params(){
		let all = document.getElementsByTagName("input");
		let collection = [];
		let htype = [];
		for(let ix = 0; ix < all.length; ++ix){
			let inp = all[ix];
			if(inp.type.toLowerCase() == "checkbox" && inp.checked ){
				if(inp.name == "collection") {
					collection.push(inp.value);
				}
				if(inp.name == "type"){
					htype.push(inp.value);
				}
			}
		}
		return({ "collection": collection, "type": htype } );
	}

	submit_search(){
		let el = document.getElementById("search.hits");
		el.innerHTML = "";
		let params = this.gather_params();

		let sb = document.getElementById('search.box');
		let obj = {
			call: 'search',
			collection: params.collection,
			type: params.type,
			term: sb.value
		};
		if(this.is_closed){
			this.setup_websocket();
		}
		this.sock.send(JSON.stringify(obj));
	}
	do_type_invert(id){
		let el = document.getElementById(id);
		let cl = el.querySelectorAll("div input");
		cl.forEach( inp => {
			inp.checked = ! inp.checked;
		});
	}

	// allow text based type selection. 
	do_filter_type(tog){
		let qin = document.getElementById("filter.type");
		let rx = new RegExp(qin.value,'i');

		let el = document.getElementById("search.params.box.type.list");
		let cl = el.querySelectorAll("div input");
		cl.forEach( inp => {
			let str = inp.value;
			if(str.match(rx)){
				inp.checked = tog;
			}else{
				inp.checked = !tog;
			}
		});
	}
	
	do_filter_collection(tog){
		let qin = document.getElementById("filter.collection");
		let rx = new RegExp(qin.value,'i');
		let el = document.getElementById("search.params.box.collection.list");
		let cl = el.querySelectorAll("div input");
		cl.forEach( inp => {
			let str = inp.value;
			if(str.match(rx)){
				inp.checked = tog;
			}else{
				inp.checked = !tog;
			}
		});
	}
	/**
	* Begin entry points:
	*
	* the main.html file calls these functions via buttons and links.
	*/
	static filter_collection(tog){
		DClient.client.do_filter_collection(tog);
	}
	static filter_type(tog){
		DClient.client.do_filter_type(tog);
	}
	static invert(what){
		switch(what) {
			case 'type':
				DClient.client.do_type_invert('search.params.box.type.list');
				break;
			case 'collection':
				DClient.client.do_type_invert('search.params.box.collection.list');
				break;
		}
	}
	static search(){
		DClient.client.submit_search();
	}
	static fetch_doc(el){
		DClient.client.do_fetch_doc(el);
	}
	static select_tab(el){
		DClient.client.do_select_tab(el);
	}
	static close_tab(el){
		DClient.client.do_close_tab(el);
	}
	static toggle_params(state){
		DClient.client.do_toggle_params();
	}
	static toggle_collection_result(al){
		DClient.client.do_toggle_collection_result(al);
	}
}

