'use strict';
const http = require('http'), url = require('url');
const path = require('path'), fs = require('fs');
const os = require('os');
const WebSocketServer = require('websocket').server;
const search = require('./searchdex.js');


function start(opt){
	opt.port = opt.port || 8080;
	if(! opt.docset){
		opt.docset = path.join(os.homedir(),".docset");
	}
	console.log("");
	console.log("Docserve startup ");
	console.log("HTTP Server listening on port: ",opt.port);
	console.log("Docset directory: ",opt.docset);
	console.log("");

	let serv = new Serve(opt);
	function handler(req, res){
		try {
			serv.dispatch(req, res);
		}catch(err){
			console.log(err);
			serv.error(req, res, err);
		}
	}
	let hs = http.createServer( handler );
	hs.listen(opt.port);
	let disp = new WSDispatch(opt,hs);
}



function serve_file(res, fn){
	let p = path.parse(fn);
	switch(p.ext){
		case '.css':
			res.writeHead(200, { 'Content-Type': 'text/css; charset=utf-8'} );
			break;
		case '.js':
			res.writeHead(200, { 'Content-Type': 'application/javascript; charset=utf-8'} );
			break;
		case '.txt':
			res.writeHead(200, { 'Content-Type': 'text/plain; charset=utf-8'} );
			break;
		case '.htm':
		case '.html':
			res.writeHead(200, { 'Content-Type': 'text/html; charset=utf-8'} );
			break;
		default:
			res.writeHead(200, { 'Content-Type': 'application/octet-stream'} );
			break;
	}
	fs.readFile(fn, 'utf8', (err, data) => {
		if(err){
			console.log(err);
			res.end();
		}else{
			res.write(data);
			res.end();
		}
	});
}
class WSDispatch {
	constructor(opt,hs){
		this.hs = hs;
		this.ds = new search.DocSet(opt.docset);
		let ws = new WebSocketServer({
			httpServer: hs,
			autoAcceptConnections: false
		});
		ws.on('request', (req) => {
			if(this.is_ws(req)){
				let con = req.accept(null,req.origin);
				this.setup_connection(con);
			}
		});
		this.ws = ws;
	}
	setup_connection(con) {
		let ws = this;
		con.on('message', msg => {
			let data = JSON.parse(msg.utf8Data);
			if(data.call){
				try {
					ws.dispatch(con, data);
				}catch(err) {
					console.log(data.call);
					console.log(err);
				}
			}
		});
		con.on('close', (code,diz) => {
			console.log(`${code} - ${diz}`);
		});
		this.do_types(con, {});
		this.do_collection(con, {});
	}
	/**
	* Dispatch the request. 
	*
	* @return {bool} True if the request was handled.
	*/
	dispatch(con, data){
		switch(data.call){
			case 'search':
				this.do_search(con,data);
				return(true);
				break;
			case 'types':
				this.do_types(con, data);
				return(true);
				break;
			default:
				console.log("Unknown: " + data.call);
				return(false);
				break;
		}
	}
	is_ws(req){
		if(req.resource == "/ws"){
			return(true);
		}
		return(false);
	}

	do_search(con,data){
		let qry = {
			term: '%' + data.term + '%'
		}
		if(data.collection){
			qry.collection = data.collection;
		}
		if(data.type){
			qry.type = data.type;
		}
		this.ds.search(qry, m => {
			con.send(JSON.stringify(m));
		});
	}
	do_types(con, _data){
		function gather( name ){
			con.send(JSON.stringify({data_type: 'id.type', type: name}));
		}
		this.ds.types(gather).then( r => {
			// don't really need to finalize a websocket, but this is when
			// the gather() function is done.
		}).catch(err => {
			console.log(err);
		});
	}
	do_collection(con, _data){
		this.ds.collection().then( rv => {
			let rec = {
				data_type: 'id.collection',
				collection: []
			};
			rv.forEach( el => {
				rec.collection.push(el.name);
			});
			con.send(JSON.stringify(rec));
		});
	}

}


class Serve {
	constructor(opt){ 
		this.srch = new search.DocSet(opt.docset);
	}
	error(req, res, err){
		res.writeHead(500,{ 'Content-Type': 'application/json' }); 
		res.write(JSON.stringify({ resp: 'error', message: 'Check server logs'  }));
		res.end();
	}
	dispatch(req, res){
		let q = url.parse(req.url);
		let p = q.pathname.split('/');
		p.shift();
		let bp = p.shift();
		switch(bp){
			case 'static':
				return(this.do_static(req,res));
				break;
			case 'page':
				return(this.do_fetch_content(req, res));
				break;
			default:
				this.do_index(req,res);
				return(true);
		}
	}

	do_index(req, res){
		let fn = path.join(__dirname,'..','asset','main.html');
		serve_file(res,fn);
	}
	do_static(req, res){
		let q = url.parse(req.url);
		let a_dir = q.pathname.split('/');
		a_dir.shift(); // remove leading ['','static']
		a_dir.shift();
		let fn = path.join(__dirname,'..','asset',a_dir.join(path.sep));
		serve_file(res, fn);
		return(true);
	}
	do_fetch_content(req, res){
		let u = url.parse(req.url,true);
		
		let page_ar = u.path.split("/");
		// skip ahead to /page/
		while(page_ar[0] != 'page'){
			page_ar.shift();
		}
		page_ar.shift(); // last 'page'

		// Vim/path/to/vim-doc.html
		let set = page_ar.shift();
		let page = page_ar.join("/");
		this.srch.fetch(set, page).then( prec => {
			let mtype = prec.mime_type;
			if( mtype.match(/^text/i) ){
				mtype += "; charset=utf-8";
			}

			let buf = prec.buffer;
			res.writeHead(200, {
				"Content-Type": mtype,
				"Content-Length": buf.length
			});
			res.write(buf.toString());
			res.end();
		}).catch(err => {
			if(err.code == 'ENOENT'){
				res.writeHead(404, { 'Content-Type': 'text/plain'} );
				res.write("Not found");
			}else{
				res.writeHead(500, { 'Content-Type': 'text/plain'} );
				res.write("Server had a boo-boo, please check logs");
				console.log(err);
			}
			res.end();
		});
		return(true);
	}
}

exports.start = start;


