'use strict';
const sqlite3 = require('sqlite3');
const path = require('path'), fs = require('fs');
const mime = require('mime/lite');

const fsp = require('fs').promises;

/**
* Maximum file size of content files.
*/
const MAX_FILE_SIZE = 67108864;

/**
* Perform the search in a specific sqlite3 database file.
*
* @param {String} dbname - sqlite3 database name. (docSet.dsidx file)
* @param {String} term  - LIKE query (must include the %'s)
* @param {Function} cb - Callback function receives each row of hits.
* @return {Promise} - resolves to the count of records found.
*/
function scan_dbf(dbname, type, term, cb){
	let db = new sqlite3.Database(dbname);
	let stype = "'" + type.join("','") + "'";

	return new Promise( (pass,fail) => {
		try {
			let stmt = db.prepare("SELECT name, type, path FROM searchIndex WHERE type IN (" + stype + ") AND name LIKE ?");
			stmt.on('error',err => { console.log("oops: ",dbname,"err: ",err); });
			stmt.each([term], (err, row) => {
				if(err){
					fail(err);
					console.log(err);
				}else{
					try {
						cb(row);
					}catch(err){
						fail(err);
					}
				}
			}, (err, count) => {
				stmt.finalize();
				if(err) {
					console.log(err);
					fail(err);
				}else{
					pass(count);
				}
			});
		}catch(err){
			fail(err);
		}
	});
}

async function get_all_types(dbname, cb){
	let db = new sqlite3.Database(dbname);
	return new Promise( (pass,fail) => {
		try {
			let stmt = db.prepare('SELECT DISTINCT type FROM searchIndex');
			stmt.on('error',err => { console.log("oops: ",dbname,"err:",err); });
			stmt.each([], (err, row) => {
				if(err){
					fail(err);
				}else{
					cb(row.type);
				}
			}, (err, count) => {
				if(err) {
					fail(err);
				}else{
					pass(count);
				}
			});
			stmt.finalize();
		}catch(err){
			fail(err);
		}
	});
}

//---- end of normal functions --

/**
* Interface for searching the Zeal/Dash "docsets". 
*/
const DocSet = function(docsets){
	this.docsets = docsets;

}

/**
* Retrieve all the types of all the collections. 
*
* @param {Function} gather - Gather results as they come in.
* 
* @return {Set} - (Promise) that Resolves to the Set of types.
*/
DocSet.prototype.types = async function(gather) {
	let col = await this.collection();
	let sent = new Set();
	if(! gather){
		gather = function() { };
	}

	let cb = function(name){
		if(! sent.has(name)) {
			sent.add(name);
			gather(name);
		}
	}
	return new Promise( (pass,fail) => {
		let prom = [];
		col.forEach( it => {
			prom.push( get_all_types(it.dbname, cb) );
		});
		Promise.all(prom).then( r => {
			pass(sent);
		}).catch(err => {
			fail(err);
		});
	});
}




DocSet.prototype.dbm_path = function(name){
	name = name.replace(/[^a-zA-Z0-9_]/g, "");
	return(path.join(this.docsets, name + ".docset", 'Contents', 'Resources', 'docSet.dsidx'));
}
DocSet.prototype.file_path = function(col,name){
	col = col.replace(/[^a-zA-Z0-9_\-\+\=\.]/g, "");
	let ix = name.lastIndexOf(">");
	if(ix > -1){
		++ix;
		name = name.substr(ix);
	}
	ix = name.indexOf("#");
	if(ix > 0){
		name = name.substr(0,ix);
	}
	return(path.join(this.docsets, col + ".docset", 'Contents', 'Resources', 'Documents', name ));
}

DocSet.prototype.fetch = function(collection, fp){
	let filename = this.file_path(collection, fp);
	let mtype = mime.getType(filename);
	let size = 0;
	return new Promise( (pass,fail) => {
		let handle = 0;
		let rbuf = null;
		fsp.open(filename).then( fh => {
			handle = fh;   // get stat info.
			return(handle.stat());
		}).then(st => {
			size = st.size;
			if(size > MAX_FILE_SIZE){
				fail("File is too large");
			}
			rbuf = Buffer.alloc(size);
			return(handle.read(rbuf,0,rbuf.length,null));
		}).then( ro => {
			handle.close();
			pass( {
				filename: filename, 
				mime_type: mtype, 
				buffer: rbuf
			});
		}).catch(err => {
			fail(err);
		});
	});
}

/**
* Search for records.
* 
* @param {Object} qry - Search info
* @param {function} cb - callback receives hits.
* @return {Promise} 
*/
DocSet.prototype.search = async function(qry, gather){
	let all_col = await this.collection();
	let want_set = null;
	const max_hits = 2000;
	let ctr = 0;

	// clean out any "'" characters. (any other non-alpha as well)
	let types = [];
	qry.type.forEach( t => {
		types.push(t.replace(/[^a-zA-Z0-9]/g,""));
	});

	// collection = ['Vim','C','JavaScript']
	if(qry.collection){
		want_set = new Set(qry.collection);
	}

	// returns a function that has access to 'col'
	function make_gather_cb(col){
		return(function(row) {
			if(++ctr < max_hits){
				row.data_type = "hit";
				row.collection = col.name;
				gather(row);
			}else{
				if(gather){
					gather( { data_type: 'hit-max', title: 'Exceeded Limit - please narrow search'} );
					gather = null;
				}
			}
		});
	}
	// want this collection?
	function wanted(name){
		if(want_set){
			return want_set.has(name);
		}else{
			return(true); // unspecified, search all.
		}
	}
	let prom = [];
	
	all_col.forEach( c => {
		if(wanted(c.name)) {	
			let p = scan_dbf(c.dbname, types, qry.term, make_gather_cb(c) );
			prom.push(p);
		}
	});
	return(Promise.all(prom));
}


/**
* Return an array of docsets
*
* @return {Promise} - Array of docset objects 
*/
DocSet.prototype.collection = async function(){
	let dir = await fs.promises.opendir(this.docsets);
	let dent = null;
	let rv = [];
	for await (dent of dir) {
		let name = path.basename(dent.name,'.docset');
		if(dent.isDirectory()){
			let dbname = path.join(this.docsets, dent.name, 'Contents/Resources/docSet.dsidx')
			let docs = path.join(this.docsets, dent.name, 'Contents/Resources/Documents');
			rv.push( {documents: docs, dbname: dbname, name: name} );
		}
	}
	return(rv);
}



exports.DocSet = DocSet;
